from tsvector import SearchVectorField
from django.db.backends.postgresql.schema import DatabaseSchemaEditor


class TriggerSchemaEditor(DatabaseSchemaEditor):

    def create_model(self, model):
        for field in model._meta.local_fields:
            if isinstance(field, SearchVectorField) and field.columns:
                self.deferred_sql.extend(self._create_tsvector_trigger(model, field))

    def delete_model(self, model):
        for field in model._meta.local_fields:
            if isinstance(field, SearchVectorField):
                self.deferred_sql.extend(self._drop_tsvector_trigger(model, field))

    def add_field(self, model, field):
        if isinstance(field, SearchVectorField) and field.columns:
            self.deferred_sql.extend(self._create_tsvector_trigger(model, field))

    def remove_field(self, model, field):
        if isinstance(field, SearchVectorField):
            self.deferred_sql.extend(self._drop_tsvector_trigger(model, field))

    def alter_field(self, model, old_field, new_field, strict=False):
        self.remove_field(model, old_field)
        self.add_field(model, new_field)

    def _tsvector_setweight(self, field):

        if field.language_column and field.language:
            language = 'COALESCE(NEW.{}::regconfig, {})'.format(
                self.quote_name(field.language_column),
                self.quote_value(field.language)
            )
        elif field.language_column:
            language = 'NEW.{}::regconfig'.format(
                self.quote_name(field.language_column)
            )
        else:
            language = self.quote_value(field.language or 'english')

        sql_setweight = (
            "setweight(to_tsvector({language}, COALESCE(NEW.{column}, '')), {weight}) ||"
        )

        weights = []
        for column in field.columns:
            weights.append(sql_setweight.format(
                language=language,
                column=self.quote_name(column.name),
                weight=self.quote_value(column.weight)
            ))
        weights[-1] = weights[-1][:-3] + ';'
        return weights

    def _tsvector_update_column_checks(self, field):
        checks = []
        for column in field.columns:
            checks.append(
                "ELSIF (NEW.{column} <> OLD.{column}) THEN do_update = true;".format(
                    column=self.quote_name(column.name)
                )
            )
        checks[0] = checks[0][3:]
        checks.append('END IF;')
        return checks

    sql_tsvector_create_trigger_function = (
        "CREATE FUNCTION {function}() RETURNS trigger AS $$\n"
        "DECLARE\n"
        " do_update bool default false;\n"
        "BEGIN\n"
        " {update_checks}\n"
        " IF do_update THEN\n"
        "  {update_body}\n"
        " END IF;\n"
        " RETURN NEW;\n"
        "END\n"
        "$$ LANGUAGE plpgsql"
    )

    def _create_tsvector_update_function(self, function, field):

        update_checks = ["do_update = true;"]
        if not field.force_update:
            update_checks = [
                                "IF (TG_OP = 'INSERT') THEN do_update = true;",
                                "ELSIF (TG_OP = 'UPDATE') THEN"] + [
                                ' ' + check for check in self._tsvector_update_column_checks(field)] + [
                                "END IF;"
                            ]

        update_body = [
                          "NEW.{} :=".format(self.quote_name(field.column))] + [
                          ' ' + weight for weight in self._tsvector_setweight(field)
                          ]

        return self.sql_tsvector_create_trigger_function.format(
            function=function,
            update_checks='\n '.join(update_checks),
            update_body='\n  '.join(update_body)
        )

    sql_create_insert_or_update_trigger = (
        "CREATE TRIGGER {trigger} BEFORE INSERT OR UPDATE"
        " ON {table} FOR EACH ROW EXECUTE PROCEDURE {function}()"
    )

    def _create_tsvector_trigger(self, model, field):

        assert field.columns, "Cannot create text search trigger without 'columns'."

        tsvector_function = self._create_index_name(model, [field.column], '_func')
        tsvector_trigger = self._create_index_name(model, [field.column], '_trig')

        yield self._create_tsvector_update_function(tsvector_function, field)

        yield self.sql_create_insert_or_update_trigger.format(
            table=self.quote_name(model._meta.db_table),
            trigger=self.quote_name(tsvector_trigger),
            function=tsvector_function,
        )

    sql_drop_trigger = "DROP TRIGGER IF EXISTS {trigger} ON {table}"
    sql_drop_trigger_function = "DROP FUNCTION IF EXISTS {function}()"

    def _drop_tsvector_trigger(self, model, field):

        tsvector_function = self._create_index_name(model, [field.column], '_func')
        tsvector_trigger = self._create_index_name(model, [field.column], '_trig')

        yield self.sql_drop_trigger.format(
            table=self.quote_name(model._meta.db_table),
            trigger=tsvector_trigger,
        )

        yield self.sql_drop_trigger_function.format(
            function=tsvector_function,
        )



from django.apps import AppConfig
from django.db.models.signals import pre_migrate


from .migrate import inject_trigger_operations


class TextSearchVectorConfig(AppConfig):
    name = 'tsvector'

    def ready(self):
        pre_migrate.connect(inject_trigger_operations)



from django.db.migrations.operations.base import Operation

from .schema import TriggerSchemaEditor


class TriggerOperation(Operation):

    def state_forwards(self, app_label, state):
        pass

    def database_forwards(self, app_label, schema_editor, from_state, to_state):
        with TriggerSchemaEditor(schema_editor.connection) as trigger_editor:
            super().database_forwards(app_label, trigger_editor, from_state, to_state)

    def database_backwards(self, app_label, schema_editor, from_state, to_state):
        with TriggerSchemaEditor(schema_editor.connection) as trigger_editor:
            super().database_backwards(app_label, trigger_editor, from_state, to_state)


def inject_trigger_operations(plan=None, **kwargs):

    if plan is None:
        return

    for migration, backward in plan:

        inserts = []
        for index, operation in enumerate(migration.operations):
            clsname, args, kwargs = operation.deconstruct()
            if clsname in ('CreateModel', 'DeleteModel', 'AddField', 'RemoveField', 'AlterField'):
                newop = type('CreateOrDropTrigger', (TriggerOperation, operation.__class__), {})
                if clsname in ('DeleteModel', 'RemoveField'):
                    # inject operation ahead of the original as we
                    # need to introspect the not-yet-deleted field
                    inserts.append((index, newop(*args, **kwargs)))
                else:
                    # inject operation after the original as we
                    # need table and column to be created first
                    inserts.append((index + 1, newop(*args, **kwargs)))

        for inserted, (index, operation) in enumerate(inserts):
            migration.operations.insert(inserted + index, operation)





from django.core import checks
from django.db.models import CharField, TextField
from django.utils.encoding import force_text
from django.utils.itercompat import is_iterable
from django.contrib.postgres.search import SearchVectorField as OriginalSearchVectorField


class WeightedColumn:

    WEIGHTS = ('A', 'B', 'C', 'D')

    def __init__(self, name, weight):
        self.name = name
        self.weight = weight

    def check(self, field, searchable_columns):
        errors = []
        errors.extend(self._check_column_name(field, searchable_columns))
        errors.extend(self._check_weight(field, self.WEIGHTS))
        return errors

    def _check_column_name(self, field, columns):
        if self.name not in columns:
            yield checks.Error(
                '{}.name "{}" is not one of the available columns ({})'.format(
                    self.__class__.__name__, self.name,
                    ', '.join(['"{}"'.format(c) for c in columns])
                ), obj=field, id='postgres.E110',
            )

    def _check_weight(self, field, weights):
        if self.weight not in weights:
            yield checks.Error(
                '{}.weight "{}" is not one of the available weights ({})'.format(
                    self.__class__.__name__, self.weight,
                    ', '.join(['"{}"'.format(w) for w in weights])
                ), obj=field, id='postgres.E111',
            )

    def deconstruct(self):
        path = "tsvector.{}".format(self.__class__.__name__)
        return path, [force_text(self.name), force_text(self.weight)], {}


class SearchVectorField(OriginalSearchVectorField):

    def __init__(self, columns=None, language=None, *args, **kwargs):
        self.columns = columns
        self.language = language
        self.language_column = kwargs.pop('language_column', None)
        self.force_update = kwargs.pop('force_update', False)
        kwargs['db_index'] = True
        kwargs['null'] = True
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        if self.columns is not None:
            kwargs['columns'] = self.columns
        if self.language is not None:
            kwargs['language'] = force_text(self.language)
        if self.language_column is not None:
            kwargs['language_column'] = force_text(self.language_column)
        if self.force_update is not False:
            kwargs['force_update'] = self.force_update
        del kwargs['db_index']
        del kwargs['null']
        return name, "tsvector.{}".format(self.__class__.__name__), args, kwargs

    def check(self, **kwargs):
        errors = super().check(**kwargs)
        textual_columns = self._find_textual_columns()
        errors.extend(self._check_columns_attribute(textual_columns))
        errors.extend(self._check_language_attributes(textual_columns))
        errors.extend(self._check_force_update_attribute())
        return errors

    def _find_textual_columns(self):
        columns = []
        # PostgreSQL trigger only has access to fields in the table, so we
        # need to make sure to exclude any fields from multi-table inheritance
        for field in self.model._meta.get_fields(include_parents=False):
            # too restrictive?
            if isinstance(field, (CharField, TextField)):
                columns.append(field.column)
        return columns

    def _check_columns_attribute(self, textual_columns):
        if not self.columns:
            return
        if not textual_columns:
            yield checks.Error(
                "No textual columns available in this model for search vector indexing.",
                obj=self, id='postgres.E100',
            )
        elif not is_iterable(self.columns) or \
                not all(isinstance(wc, WeightedColumn) for wc in self.columns):
            yield checks.Error(
                "'columns' must be an iterable containing WeightedColumn instances",
                obj=self, id='postgres.E101',
            )
        else:
            for column in self.columns:
                for error in column.check(self, textual_columns):
                    yield error

    def _check_language_attributes(self, textual_columns):
        if self.columns and not any((self.language, self.language_column)):
            yield checks.Error(
                "'language' or 'language_column' is required when 'columns' is provided",
                obj=self, id='postgres.E102',
            )
            return
        if self.language and not isinstance(self.language, str):
            # can we get list of available langauges?
            yield checks.Error(
                "'language' must be a valid language",
                obj=self, id='postgres.E103',
            )
        if self.language_column and self.language_column not in textual_columns:
            yield checks.Error(
                """'language_column' "{}" is not one of the available columns ({})""".format(
                    self.name, ', '.join(['"{}"'.format(c) for c in textual_columns])
                ), obj=self, id='postgres.E104',
            )

    def _check_force_update_attribute(self):
        if self.force_update not in (None, True, False):
            yield checks.Error(
                "'force_update' must be None, True or False.",
                obj=self, id='postgres.E105',
            )
